/*
******************************************************************************
*
*      File             : tfo_decision.c
*      Version          : 2.0.0
*      Date             : 27 February 2001
*      Author           : Jochen Junkawitsch
*      Description      : source file for TFO decision algorithm
*
*
******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include "tfo_decision.h"







/* ---------------------------------------------------------------------------------
                     Main TFO Decision Procedure (section 11.1)
   --------------------------------------------------------------------------------- */

int tfo_decision_main(Data* data)
{
  /* Two compatible AMR codec types? */
  if (data->lType == UMTS_AMR_2 && data->dType == UMTS_AMR_2 ||
	  data->lType == UMTS_AMR   && data->dType == UMTS_AMR   ||
      data->lType == FR_AMR     && data->dType == FR_AMR     ||
      data->lType == HR_AMR     && data->dType == HR_AMR     ||
      data->lType == FR_AMR     && data->dType == HR_AMR     ||
      data->lType == HR_AMR     && data->dType == FR_AMR     ||
      data->lType == UMTS_AMR_2 && data->dType == HR_AMR     ||
      data->lType == HR_AMR     && data->dType == UMTS_AMR_2 ||
      data->lType == UMTS_AMR_2 && data->dType == FR_AMR     ||
      data->lType == FR_AMR     && data->dType == UMTS_AMR_2 ||
      data->lType == UMTS_AMR_2 && data->dType == UMTS_AMR   ||
      data->lType == UMTS_AMR   && data->dType == UMTS_AMR_2   )
  {
    return tfo_decision_AMR(data);
  }


  /* Same non AMR codec type? */
  if (data->lType == GSM_FR  && data->dType == GSM_FR  ||
      data->lType == GSM_EFR && data->dType == GSM_EFR ||
      data->lType == GSM_HR  && data->dType == GSM_HR    )
  {
    return immediate_TFO(data);
  }

  /* Codec Mismatch Resolution */
  return codec_mismatch_resolution(data);

}


/* ---------------------------------------------------------------------------------
                TFO Decision Algorithm for AMR Codec Types (section 11.2)
   --------------------------------------------------------------------------------- */


int tfo_decision_AMR (Data* data)
{
  /* LACS == DACS ? */
  if (data->lacs == data->dacs)
    return immediate_TFO(data);

  /* FR - HR - Matching ? */
  if (FR_HR_matching(data))
  {
    immediate_TFO(data);
    return FR_HR_optimization(data);
  }

  /* OACS acceptable ? */
  if (!acceptable(data, data->oacs))
    return codec_mismatch_resolution(data);

  /* IACS == OACS ? */
  if (data->cacs == data->oacs)
    return immediate_TFO(data);

  /* IACS subset of OACS ? */
  if (contiguous_subset(data->cacs, data->oacs))
  {
    immediate_TFO(data);
    return codec_mode_optimization(data);
  }

  return change_to_OACS(data);
}


/* ---------------------------------------------------------------------------------
                          FR - HR - Matching (section 12.6)
   --------------------------------------------------------------------------------- */

int FR_HR_matching(Data* data)
{
  if (!(data->lType == FR_AMR     && data->dType == HR_AMR ||
		data->lType == HR_AMR     && data->dType == FR_AMR ||
		data->lType == UMTS_AMR_2 && data->dType == HR_AMR ||
		data->lType == HR_AMR     && data->dType == UMTS_AMR_2))
    return 0;

  if (!data->cacs)
    return 0;

  if (!acceptable(data, data->cacs))
    return 0;

  printf("FR-HR-Matching.\n");
  return 1;
}

/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

int determine_IACS(Data* data)
{
  /* calculate CSCS */
  data->cscs = data->lscs & data->dscs;

  /* calculate CMACS */
  data->cmacs = data->lmacs < data->dmacs ? data->lmacs : data->dmacs;

  /* calculate IACS */
  data->cacs = contiguous_subset_selection((CodecSet)(data->lacs & data->dacs), 0,0,
					   data->lacs, data->dacs, data->cmacs);
  return 0;
}

/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

int determine_OACS(Data* data)
{
  /* no side supports ACS change */
  if (!data->lOM && !data->dOM)
    calc_OACS_case1(data);

  /* both sides support ACS change */
  else if (data->lOM && data->dOM)
    /* calculate OACS */
    data->oacs = calc_OACS(data->cscs,
			   data->cmacs,
			   data->lType == HR_AMR || data->dType == HR_AMR ? HR_AMR : FR_AMR);

  /* only one side supports ACS change */
  else
    calc_OACS_case2(data);

  return 0;
}

/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

int acceptable(Data* data, CodecSet cs)
{
  CodecType codecType = data->lType == HR_AMR || data->dType == HR_AMR ? HR_AMR : FR_AMR;
  return lowModeRule (data->lacs, data->dacs, cs, codecType) &&
         highModeRule(data->lacs, data->dacs, cs);;
}

/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

/* check if a is a contiguous subset of b */
int contiguous_subset(CodecSet a, CodecSet b)
{
  int mode;

  if (!a) return 0;

  for (mode=mode_122; mode; mode>>=1)
  {
    if (a & mode) break;
    b &= mode ^ 0xFF;
  }

  return a == b;
}


/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

/* section 12.2.2 Case 1: No side supports ACS change */
int calc_OACS_case1(Data* data)
{
  CodecSet cacs = data->lacs & data->dacs;
  if (!contiguous_subset(cacs, data->lacs) || !contiguous_subset(cacs, data->dacs))
  {
    data->oacs = contiguous_subset_selection(cacs, data->lOM, data->dOM,
					     data->lacs, data->dacs, data->cmacs);
    return 0;
  }
  data->oacs = cacs;
  return 0;
}


/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

/* section 12.2.2 Case 2: Only one side supports ACS change */
int calc_OACS_case2(Data* data)
{
  CodecSet fixACS, scs, cscs;

  if (data->lOM && !data->dOM)
  {
    fixACS = data->dacs;
    scs = data->lscs;
  }
  else if (!data->lOM && data->dOM)
  {
    fixACS = data->lacs;
    scs = data->dscs;
  }
  else
  {
    fprintf(stderr, "ERROR in calc_OACS_case2()\n");
    exit(0);
  }

  cscs = fixACS & scs;
  
  if (!contiguous_subset(cscs, fixACS))
  {
    data->oacs = contiguous_subset_selection(cscs, data->lOM, data->dOM,
					     data->lacs, data->dacs, data->cmacs);
    return 0;
  }

  if (nModes(cscs) > data->cmacs)
  {
    data->oacs = contiguous_subset_selection(cscs, data->lOM, data->dOM,
					     data->lacs, data->dacs, data->cmacs);
    return 0;
  }

  data->oacs = cscs;

  return 0;
}


/* ---------------------------------------------------------------------------------
                        
   --------------------------------------------------------------------------------- */

/* section 12.4 Rules for contiguous subset selection */
CodecSet contiguous_subset_selection (CodecSet cs, int lOM, int dOM,
				      CodecSet lacs, CodecSet dacs, int cmacs)
{
  int mode;
  for (mode=mode_122; mode; mode>>=1)
  {
    if ( (lOM || contiguous_subset(cs, lacs)) &&
	 (dOM || contiguous_subset(cs, dacs)) && 
	 (nModes(cs) <= cmacs))
      break;

    cs &= mode ^ 0xFF;
  }
  return cs;
}



/* ---------------------------------------------------------------------------------
                                  terminal blocks
   --------------------------------------------------------------------------------- */

int immediate_TFO(Data* data)
{
  if (data->lType == FR_AMR || data->lType == HR_AMR ||
	  data->lType == UMTS_AMR || data->lType == UMTS_AMR_2)
	printf("Immediate TFO on IACS.\n");
  else
	printf("Immediate TFO on %s.\n", codecName(data->lType));
  return 0;
}

int codec_mismatch_resolution(Data* data)
{
  printf("Codec Mismatch Resolution.\n");
  return 0;
}

int change_to_OACS(Data* data)
{
  printf("Change ACS to OACS and establish TFO on OACS.\n");
  return 0;
}

int codec_mode_optimization(Data* data)
{
  printf("Codec Mode Optimization.\n");
  return 0;
}

int FR_HR_optimization(Data* data)
{
  printf("FR - HR - Optimization.\n");
  return 0;
}

/* ---------------------------------------------------------------------------------
                                  help routines
   --------------------------------------------------------------------------------- */

/* return a string with the name of a mode */
char* modeName(CodecMode m)
{
  switch(m)
  {
	case mode_122: return "12.2";
	case mode_102: return "10.2";
	case mode_795: return "7.95";
	case mode_740: return "7.40";
	case mode_670: return "6.70";
	case mode_590: return "5.90";
	case mode_515: return "5.15";
	case mode_475: return "4.75";
	default:
	  fprintf(stderr, "ERROR in modeName()\n");
	  exit(0);
  }
  return 0;
}

	  

/* return a string with the name of a codec type */
char* codecName(CodecType t)
{
  switch(t)
  {
	case FR_AMR:     return "FR_AMR";
	case HR_AMR:     return "HR_AMR";
	case UMTS_AMR:   return "UMTS_AMR";
	case UMTS_AMR_2: return "UMTS_AMR_2";
	case GSM_EFR:    return "GSM_EFR";
	case GSM_FR:     return "GSM_FR";
	case GSM_HR:     return "GSM_HR";
	default:
	  fprintf(stderr, "ERROR in modeName()\n");
	  exit(0);
  }
  return 0;
}

